from grum.rpc import RPCClient
from bec_utils import BECMessage, MessageEndpoints
from bec_utils import RedisConnector
from os import listdir
from os.path import isfile, join
import h5py
import time
import numpy as np
from PIL import Image

class BECAdapter:

    def __init__(self, redis_connector, recon = False, monitor=False, motor_pos = False) -> None:
        super().__init__()

        host = "localhost"
        port = 8000
        self.redis_connector = redis_connector
        self.eaccount = "e20631" #change for each new beamtime

        # To decide what should be plotted
        self.monitor = monitor
        self.motor_pos = motor_pos
        self.recon = recon
        
        if self.recon:
            self.grum_client_recon_abs = RPCClient(host, port)  
            self.grum_client_recon_phase = RPCClient(host, port)   
            self.last_plotted_scan_nr = 0
            self.latest_reconstructed = 5000        

        if self.monitor:
            self.grum_client_monitor = RPCClient(host ,port)
            self.grum_client_monitor.new_plot("monitor",{'xlabel':'time', 'ylabel':"sls_ring_current", "time": True, "frmt":"Line" })

        if self.motor_pos:
            self.grum_client_motor_pos_static = RPCClient(host, port)  
            self.grum_client_motor_pos_moving = RPCClient(host, port)   
            self.scan_nr = None 
            self.scan_running = False
            self.grum_client_motor_pos_static.new_plot("Motor position static", {'xlabel':'average_x_st_fzp', 'ylabel':"average_y_st_fzp" , "frmt": "Dots"})
            self.grum_client_motor_pos_moving.new_plot("Motor position move", {'xlabel':'average_x_st_fzp', 'ylabel':"average_y_st_fzp" , 'frmt':"Motor pos move"})

    def start(self):
        if self.motor_pos:
            self.start_scan_status_sub()
            self.start_scan_segment_sub()
        if self.monitor:
            self.start_monitor_sub()
        
    def start_scan_status_sub(self):
        self._scan_sub = self.redis_connector.consumer(
            MessageEndpoints.scan_status(), cb=self.scan_status_update
        )
        self._scan_sub.start()
    
    def scan_status_update(self, segment):
        # whenever a new scan starts or ends, for now changes scan nr both in start and end
        segment = BECMessage.ScanStatusMessage.loads(segment.value)
        if segment.content["status"] == "open":
            self.scan_nr = segment.content["info"]["scan_number"]
            self.positions = segment.content["info"]["positions"]
            self.grum_client_motor_pos_moving.new_plot("Motor position move", {"title":"scan number:" + str(self.scan_nr),'xlabel':'average_x_st_fzp', 'ylabel':"average_y_st_fzp" })
            xpos = [x for x,y in self.positions]
            ypos = [y for x,y in self.positions]
            if len(segment.content["info"]["primary"]) == 2:
                self.scan_running = True
                if self.motor_pos:
                    print("new plot")
                    # for point in self.positions
                    # self.grum_client_motor_pos_static.append_data("Motor position static", self.positions)
                    self.grum_client_motor_pos_static.new_plot( "Motor position static", { "xs": xpos, "ys":ypos, 'xlabel':'average_x_st_fzp', 'ylabel':"average_y_st_fzp", "frmt": "Dots"})
                    #Trying to update plots with "open new plots" unticked 
                    # self.grum_client_motor_pos_static.new_plot( "Motor position static", {"xs": xpos[:-1], "ys":ypos[:-1], 'xlabel':'average_x_st_fzp', 'ylabel':"average_y_st_fzp", "frmt": "Dots"})
                    # self.grum_client_motor_pos_static.append_data( "Motor position static", [xpos[-1], ypos[-1]] )

            else:
                self.scan_running = False

            if len(segment.content["info"]["primary"]) == 1:
                print("umv")
        
    def start_scan_segment_sub(self):
        self._scan_segment_sub = self.redis_connector.consumer(
            MessageEndpoints.scan_segment(), cb=self.scan_segment_update
        )
        self._scan_segment_sub.start()


    def scan_segment_update(self, segment):
        if self.scan_running:
            segment = BECMessage.ScanMessage.loads(segment.value)
            x_val = segment[0].content["data"]["average_x_st_fzp"]["value"]
            y_val = segment[0].content["data"]["average_y_st_fzp"]["value"]

            point = [x_val, y_val]
            if self.motor_pos:
                self.grum_client_motor_pos_moving.append_data("Motor position move", point)   

    def start_monitor_sub(self):
        self._monitor_sub = self.redis_connector.consumer(
            pattern = MessageEndpoints.device_readback('*'), 
            cb=self.monitor_update
        )
        self._monitor_sub.start()

    def monitor_update(self, msg):
        # Whenever we get a new monitor value
        dev = msg.topic.decode().split(MessageEndpoints._device_readback + '/')[-1].split(':sub')[0]
        msg = BECMessage.DeviceMessage.loads(msg.value)
        if dev == "sls_ring_current":
            x = msg.content["signals"][dev]['timestamp']
            y = msg.content["signals"][dev]['value']
            point = [x,y]
            if self.monitor:
                self.grum_client_monitor.append_data( "monitor", point)

    def plot_img_from_h5(self):
        
        while True:
            print('Check for new reconstruction')
            self.update_latest_reconstructed()
            print('latest_reconstructed: ', self.latest_reconstructed)
            print("last plotted scan nr: ", self.last_plotted_scan_nr)
            
            if self.latest_reconstructed > self.last_plotted_scan_nr:
                ending_path = self._get_scan_dir(1000, self.latest_reconstructed, leading_zeros = 5)
                mypath = "/sls/X12SA/data/" + self.eaccount + "/Data10/analysis/" + ending_path           
                recon_file = self.get_recon_file(mypath)
                with h5py.File(mypath + '/' + recon_file, 'r') as hf:
                    recon_img_abs, recon_img_phase = self.get_recon_imgs(hf)
                print('plotting new image with scan_nr: ', self.latest_reconstructed)
                xlab = str(self.latest_reconstructed)
                if self.recon:
                    self.grum_client_recon_abs.new_image( "Absorption", {'image': recon_img_abs, "xlabel": xlab}) #, "levels":[0.9, 1.1]})
                    # self.grum_client_recon_phase.new_image( f"Phase for scan_nr: {self.latest_reconstructed}", {'image': recon_img_angle, "xlabel": xlab}) #, "colormap":"CET-C1"})
                    self.grum_client_recon_phase.new_image( "Phase", {'image': recon_img_phase, "xlabel": xlab})#, "levels": [-1,1]}) #, "colormap":"CET-C1"})
                self.last_plotted_scan_nr = self.latest_reconstructed

            time.sleep(10)
    def get_recon_file(self, mypath):   
        files = [f for f in listdir(mypath) if isfile(join(mypath, f))]
        for file in files:
            if file.endswith("recons.h5"):
                return file
        return None
    
    def get_recon_imgs(self, hf):   
        recon_object = hf.get('reconstruction/object')
        # recon_probes = hf.get('reconstruction/probes')

        recon_object =np.array(recon_object)
        recon_object_abs = np.abs(recon_object)
        recon_object_phase = np.angle(recon_object)
        
        recon_object_abs = np.rot90(recon_object_abs,3)
        recon_object_phase = np.rot90(recon_object_phase, 3)

        recon_img_abs = recon_object_abs.tolist()
        recon_img_phase = recon_object_phase.tolist()
        return recon_img_abs, recon_img_phase
   
    def update_latest_reconstructed(self):
        # mypath = '/sls/X12SA/Data10/' + self.eaccount + '/analysis/online/ptycho/gallery'
        mypath = '/sls/X12SA/data/' +self.eaccount+ '/Data10/specES1/ptycho_reconstruct/done' # ptycho_reconstruct argument?
        file_list = listdir(mypath)
        scans = [file_name.split('_')[1] for file_name in file_list] #0 if you do gallery, 1 if you do done
        self.latest_reconstructed = max([int(scan[0:5]) for scan in scans]) # 1:6 for gallery, 0:5 for done

    def _get_scan_dir(self, scan_bundle, scan_number, leading_zeros=None):
        if leading_zeros is None:
            leading_zeros = len(str(scan_bundle))
        floor_dir = scan_number // scan_bundle * scan_bundle
        return f"S{floor_dir:0{leading_zeros}d}-{floor_dir+scan_bundle-1:0{leading_zeros}d}/S{scan_number:0{leading_zeros}d}"

if __name__ == "__main__":
    print('starts BECAdapter')
    print("initializing redis-connector")
    redis_url = "129.129.122.75:6379"  # for LamNI
    redis_connector = RedisConnector(redis_url)
    ba = BECAdapter(redis_connector, recon=True) #, motor_pos=True)  #, monitor=True)
    ba.start()
    ba.plot_img_from_h5()
